#!/bin/bash
# Purpose: _Fully_ rip a DVD or BD to MKV (DVD only for now).
#
# LICENCE: GNU GPL v3 or later
#
# Dependencies (needed in PATH):
# — isosize (part of util-linux)                    : get sector size and count of the disc (unless: -n)
# — dd, ddrescue                                    : copy image of disc                    (unless: -n)
# — lsdvd                                           : get DVD metadata
# — mplayer                                         : rip raw DVD streams in case ffmpeg is unable to rip
# — ffmpeg, ffprobe (part of FFMpeg)                : probe, rip and encode to HEVC/MKV
# — spumux (part of dvdauthor)                      : generate missing VOBSUB streams to align all tracks
# — identify, montage, convert (part of ImageMagick): get extracted frame metadata, create visual timelines
# — czkawka_cli (part of Czkawka)                   : compare timelines and detect duplicate chapters
# — mkvmerge (part of MKVToolNix)                   : get MKV metadata and assemble chapters into a final MKV
# — sed, gawk, grep, sort, uniq, head, cut, tr, wc  : data manipulation
# — dirname, tee, ln, mktemp, cat, touch, mv, rm    : file manipulation
# — jq                                              : for parsing JSON and extracting metadata
# — bc                                              : arbitrary precision calculator
# — date                                            : conversion between timestamps and hh:mm:ss.nnnnnnnnn
#
# Options:
# [-h                  : show this help                                                                         ]
# [-?                  : display metadata and exit (implies: -n -v quiet)                                       ]
# [-l lang1,lang2,…    : all languages to retain for audio and subtitles, with preferred first; default: fre,eng]
# [-v debug|info|quiet : verbosity; default: info                                                               ]
# [-t temp_path        : location with at least 2×disc size, to store temporary files; default: /tmp/bdvdrip-xxx]
# [-i input_path       : path of the disc device or ISO; default: /dev/sr0                                      ]
# [-o output_path      : path of target video file to write (`.mkv` extension); default: ./output.mkv           ]
# [-n                  : do Not dump the input_path to an ISO file; symlink instead                             ]
# [-g MAX_WxMAX_H      : maximum geometry of the encoded video (after border removal); default: 1280x720 (720p) ]
#
# IMPORTANT: https://unix.stackexchange.com/a/493393
#
# Technical references:
# — http://mod16.org/hurfdurf/?p=8
# — https://www.reddit.com/r/mkvtoolnix/comments/11nbfy0/tutorial_mediumlinked_tiny_segmented_mkvs_with/
# — https://www.reddit.com/r/handbrake/comments/bhqxve/handbrake_settings_explained/

trap 'exit 100' SIGINT

# read command-line arguments

while [ $# -gt 0 ]; do
  case "$1" in
  -h) sed -nr '2,/^$/s/^# ?//p' "$0"; exit 0 ;;
  -\?) NO_DD=true; DEBUG=quiet; SHOW=true ;;
  -l) ALL_LANGS="$2"; shift ;;
  -v) DEBUG="$2"; shift ;;
  -t) BUILD="$2"; shift ;;
  -i) DVD="$2"; shift ;;
  -o) TARGET="$2"; shift ;;
  -n) NO_DD=true ;;
  -g) IFS=x read -r MAXW MAXH <<<"$2"; MAXW=$((MAXW/16*16)); MAXH=$((MAXH/16*16)) shift ;;
  *) { echo "Unknown option: $1"; sed -n '2,/^$/s/^..//p' "$0"; } >&2; exit 1 ;;
  esac
  shift
done

ALL_LANGS=${ALL_LANGS:-fre,eng}
DEFAULT_LANG=${ALL_LANGS%%,*}
KEEP_ALL=${BUILD:+true}
BUILD="${BUILD:-$(mktemp --tmpdir -d bdvdrip-XXXXXX)}"
[ -d "$BUILD" ] && [ -w "$BUILD" ] || { echo "$BUILD is not a writeable directory"; exit 1; }
DVD="${DVD:-/dev/sr0}"
[ -r "$DVD" ] && ! [ -d "$DVD" ] || [ -e "$BUILD/.iso" ] || { echo "$DVD is not a useable readable input"; exit 1; }
TARGET="${TARGET:-./output.mkv}"
[ -d "$(dirname "$TARGET")" ] && [ -w "$(dirname "$TARGET")" ] || { echo "$(dirname "$TARGET") is not a writeable directory"; exit 1; }
if ! [[ "$MAXW" =~ ^[0-9]+$ ]]; then MAXW=1280; fi
if ! [[ "$MAXH" =~ ^[0-9]+$ ]]; then MAXH=720; fi

exec 3>&2
function log_and_run() {
  local getOutput output O E excode
  if [ "$1" == '-o' ]; then
    getOutput=1; shift
  fi
  local step="$1"; shift
  if [ "$DEBUG" != quiet ]; then
    printf '\n>> %s\n>> %s\n' "$step" "${*@Q}" >&3
  fi
  case "$getOutput.$DEBUG" in
  1.debug) "$@" 1> >(tee >(cat >&3)) 2> >(tee >(cat >&3) >&2) ;;
  1.*)     O="$(mktemp --tmpdir XXX.out)"; E="$(mktemp --tmpdir XXX.err)"; "$@" 1>"$O" 2>"$E"; excode=$?; output="$(cat "$O" "$E")"; cat "$O"; cat "$E" >&2; rm -f "$O" "$E" ;;
  .debug)  "$@" 1>&3 2>&3 ;;
  .*)      output="$("$@" 2>&1)" ;;
  esac
  excode=${excode:-$?}
  if [ $excode -ne 0 ] && [ "$DEBUG" != quiet ]; then
    [ -n "$output" ] && echo "$output"
    printf '>>> ERROR: EXIT CODE %s\n\n' "$excode" >&3
  fi
  return $excode
}

# copy DVD

if ! [ -f "$BUILD/.iso" ]; then
  if [ -n "$NO_DD" ]; then
    RM_ISO=${SHOW:+true}
    ln -s "$DVD" "$BUILD/.iso"
  else
    read -r sectcount sectsize < <(LANG=C isosize -x "$DVD" | sed -r 's/.*: (.*), .*: (.*)$/\1 \2/')

    log_and_run 'Dump disc to ISO image' \
    dd bs=$sectsize count=$sectcount if="$DVD" of="$BUILD/.iso" || {
      log_and_run 'Dump disc to ISO image (slower because of media errors)' \
      ddrescue -n -b$sectsize "$DVD" "$BUILD/.iso" "$BUILD/.iso.mapfile"
      log_and_run 'Try to recover damaged sectors if needed' \
      ddrescue -d -r1 -b$sectsize "$DVD" "$BUILD/.iso" "$BUILD/.iso.mapfile"
    }
  fi
fi

if ! [ -f "$BUILD/.json" ]; then
  meta='{}'
  for tnum in $(
    log_and_run -o 'Read disc track list' \
    lsdvd -Ox "$BUILD/.iso" | grep -o '<ix>[^<]</ix>' | tr -dc '0-9\n'
  ); do
    tmeta='{}'

    # general
    data="$(
      log_and_run -o "Read track $tnum general metadata" \
      ffprobe -hide_banner -output_format json -show_format -f dvdvideo -trim false -title $tnum "$BUILD/.iso" 2>/dev/null \
      | jq -r "select(.format?.duration? | tonumber? > 5)"
    )"
    [ -n "$data" ] && x="$(jq --argjson D "$data" '.format = $D.format' <<<"$tmeta")" && tmeta="$x" || continue

    # video
    data="$(
      log_and_run -o "Read track $tnum video metadata" \
      ffprobe -hide_banner -output_format json -show_streams -select_streams V -f dvdvideo -trim false -title $tnum "$BUILD/.iso" 2>/dev/null | jq
    )"
    [[ "$data" =~ [[:blank:]]\{ ]] && x="$(jq --argjson D "$data" '.video = $D.streams[0]' <<<"$tmeta")" && tmeta="$x" || continue

    # chapters
    # ⚠ With `-preindex 1`: chapter list may be truncated (end is missing)
    #  without `-preindex 1`: chapter list may be empty (all is missing)
    # ⚠ Chapters <1s do not get merged correctly in the end.
    # ⚠ Single-chapter titles may appear as having none.
    chapters="$(
      log_and_run -o "Merge two versions of track $tnum chapters metadata" \
      jq -sr '{chapters: [[.[] | if (.chapters|length == 0) then {} else [.chapters[] | {"\(.id)": .}] | add end] | add | to_entries[] | select((.value.end_time|tonumber)-(.value.start_time|tonumber)>1) | .value]}' \
        <(log_and_run -o "Read track $tnum chapters metadata (with \`-preindex 1\`)" \
          ffprobe -hide_banner -output_format json -show_chapters -f dvdvideo -preindex 1 -trim false -title $tnum "$BUILD/.iso" 2>/dev/null) \
        <(log_and_run -o "Read track $tnum chapters metadata (without \`-preindex 1\`)" \
          ffprobe -hide_banner -output_format json -show_chapters -f dvdvideo -trim false -title $tnum "$BUILD/.iso" 2>/dev/null)
    )"
    [[ "$chapters" =~ \"chapters\":\ (null|\[\]) ]] && chapters="$(
      jq --argjson T "$tmeta" '{chapters: [{id: 0, start_time: "0.000000", end_time: $T.format.duration}]}' <<<"{}"
    )"
    while read index chnum startpos; do
      if ! [ -f "$BUILD/.t.$tnum.ch.$chnum.sample.mpeg" ]; then
        log_and_run "Extract a sample of title $tnum chapter $chnum" \
        mplayer dvd://$tnum -chapter $chnum-$chnum -endpos $(bc -lq <<<"6+$startpos") -dumpstream -dumpfile "$BUILD/.t.$tnum.ch.$chnum.sample.mpeg" -dvd-device "$BUILD/.iso" </dev/null
      fi
      for t in a:audio s:subtitles; do
        if ffprobe -hide_banner -output_format json -show_streams -select_streams ${t%:*} "$BUILD/.t.$tnum.ch.$chnum.sample.mpeg" \
        </dev/null 2>/dev/null | jq | grep -qF '[]'; then
          data="$(
            ffprobe -hide_banner -output_format json -show_streams -select_streams ${t%:*} -f dvdvideo -trim false -title $tnum -chapter_start $chnum -chapter_end $chnum "$BUILD/.iso" 2>/dev/null \
            | jq -r 'if (.streams? == null) then [] else [.streams[] | select(.codec_type != "audio" or .bit_rate != null)] end'
          )"
        else
          ref="$(
            ffprobe -hide_banner -output_format json -show_streams -select_streams ${t%:*} -f dvdvideo -trim false -title $tnum -chapter_start $chnum -chapter_end $chnum "$BUILD/.iso" </dev/null 2>/dev/null \
            | jq -r '[.streams[] | {key: .id, value: {idx: .index, lng: (if (.tags?.language == null) then "und" else .tags.language end)}}] | from_entries'
          )"
          data="$(
            log_and_run -o "Probe track $tnum chapter $chnum ${t#*:} metadata from sample" \
            ffprobe -hide_banner -output_format json -show_streams -select_streams ${t%:*} "$BUILD/.t.$tnum.ch.$chnum.sample.mpeg" </dev/null 2>/dev/null \
            | jq -r --argjson R "$ref" "\
              if (.streams? == null) then [] else [ \
                .streams[] | \$R.[\"\\(.id)\"] as \$ref \
                | select(\$ref != null and (.codec_type != \"audio\" or .bit_rate != null)) \
                | .index = \$ref.idx \
                | .tags = {language: \$ref.lng} \
              ] end"
          )"
        fi
        x="$(jq --argjson D "$data" ".chapters[$index].${t#*:} = \$D" <<<"$chapters")" && chapters="$x"
      done
    done < <(jq -r '.chapters | to_entries[] | "\(.key) \(.value.id+1) \(.value.start_time)"' <<<"$chapters")
    x="$(jq --argjson C "$chapters" '.chapters = $C.chapters' <<<"$tmeta")" && tmeta="$x"
    x="$(jq --arg N $tnum --argjson T "$tmeta" '.["\($N)"] = $T' <<<"$meta")" && meta="$x"
  done
  # filter-out titles with no audio
  jq -r '[to_entries[] | select(([.value.chapters[].audio[]] | length) > 0)] | from_entries' <<<"$meta" >"$BUILD/.json"

  # detect crop
  list="$(jq -r 'to_entries[] | "\(.key) \((.value.format.duration / ".")[0])"' <"$BUILD/.json")"
  while read tnum vseconds; do
    case $vseconds in
    ?|??) start=0; duration=$vseconds ;;
    *)    start=$(($vseconds/3)); duration=60 ;;
    esac
    x="$(jq ".\"$tnum\".video.crop = \"$(
      log_and_run -o "Detect track $tnum picture borders" \
      ffmpeg -hide_banner -f dvdvideo -preindex 1 -trim false -title $tnum -i "$BUILD/.iso" \
        -ss $start -t $duration -vf cropdetect=limit=24:round=2:reset=0:skip=0,metadata=mode=print -f null - </dev/null 2>&1 \
      | grep -o 'crop=[0-9:]*' | sort | uniq -c | sort -k1,1nr | head -n 1 | sed 's/.*=//'
    )\"" <"$BUILD/.json")" \
    && printf '%s' "$x" >"$BUILD/.json"
  done <<<"$list"
fi

ALL_META="$(
  log_and_run -o 'Parse metadata' \
  jq -r "to_entries[] | .key as \$tnum | .value | \
    \"TITLE \(\$tnum) (\(.format.duration)s)\", \
    \"\tVideo (index, codec, geometry, frame-rate, chroma-subsampling, pixel aspect ratio, crop width:height:left:top, is interlaced, id)\", \
    (.video | \
      \"V-\(\$tnum)\t\t\(.index)\t\(.codec_name)\t\(.width)x\(.height)\t\(.avg_frame_rate)\t\(.pix_fmt)\t\(.sample_aspect_ratio | sub(\":\"; \"/\"))\t\(.crop)\t\(\",\(.field_order),\" | test(\",(bt|tb|tt),\"))\t\(.id)\"), \
    \"\tChapters (index, begin-end timestamps, name)\", \
    (.chapters[] | (.id+1) as \$chnum | ( \
      \"C-\(\$tnum)\t\(\$chnum)\t\t\(.start_time)s-\(.end_time)s\tCH.\(\$chnum)\", \
      (select (.audio|length > 0) | \"\t  Audio (index, language, codec, channels, layout, sample-rate, bit/s, is default, is commentary, is secondary, for visually impaired, id, description)\"), \
      (.audio[] | \
        \"A-\(\$tnum)\t\(\$chnum)\t\(.index)\t\(if (.tags?.language == null) then \"und\" else .tags.language end)\t\(.codec_name)\t\(.channels)\t\(.channel_layout)\t\(.sample_rate)\t\(.bit_rate)\t\(.disposition.default != 0)\t\(.disposition.comment != 0 or .disposition.descriptions != 0)\tfalse\t\(.disposition.visual_impaired != 0)\t\(.id)\t\(if (.tags?.language? != null) then \"\(.tags.language | ascii_upcase) \" else \"UND \" end)\(.channel_layout)\"), \
      (select (.subtitles|length > 0) | \"\t  Subtitles (index, language, codec, is default, is commentary, is forced, for hearing impaired, id, description)\"), \
      (.subtitles[] | \
        \"S-\(\$tnum)\t\(\$chnum)\t\(.index)\t\(if (.tags?.language == null) then \"und\" else .tags.language end)\t\(.codec_name)\t\(.disposition.default != 0)\t\(.disposition.comment != 0 or .disposition.descriptions != 0)\t\(.disposition.forced != 0)\t\(.disposition.hearing_impaired != 0 or .disposition.captions != 0)\t\(.id)\t\(if (.tags?.language? != null) then .tags.language | ascii_upcase else \"UND\" end)\"), \
      \"-\" \
    ))" \
    <"$BUILD/.json"
)"

# show information then exit, if requested

if [ -n "$SHOW" ]; then
  echo "$ALL_META"
  [ -z "$KEEP_ALL" ] && rm -rf "$BUILD" || { [ -n "$RM_ISO" ] && rm -f "$BUILD/.iso"; }
  exit 0
fi

# parse general information

declare TITLE_LIST=$(sed -rn 's/^TITLE (.*) .*/\1/p' <<<"$ALL_META")
declare VFORMAT=$(awk -F$'\t' '$5~"x480$"{print "ntsc";exit}' <<<"$ALL_META")
declare -A SRC_VIDEO
declare -A VSECONDS
declare -A CHAPTERS
declare -A CSECONDS
declare -A CHNAMES

while read -r tnum vnum; do SRC_VIDEO[$tnum]="$vnum"; done < <(
  sed -nr 's/^V-([0-9]+)\t\t([0-9]+)\t.*/\1 \2/p' <<<"$ALL_META")
while read -r tnum s; do VSECONDS[$tnum]=$s; done < <(
  sed -nr 's/^TITLE (.*) \(([0-9]+)(\..*)?s\)/\1 \2/p' <<<"$ALL_META")
for tnum in $TITLE_LIST; do CHAPTERS[$tnum]=$(
  sed -rn "s/^C-$tnum\\t([^\\t]*)\\t.*/\\1/p" <<<"$ALL_META"); done
while read -r tnum chnum from to; do CSECONDS[$tnum.$chnum]=$(bc -lq <<<"$to-$from"); done < <(
  sed -nr 's/^C-([0-9]+)\t([0-9]+)\t\t([0-9.]*)s-([0-9.]*)s\t.*/\1 \2 \3 \4/p' <<<"$ALL_META")
while read -r tnum chnum name; do CHNAMES[$tnum.$chnum]="$name"; done < <(
  sed -nr 's/^C-([0-9]+)\t([0-9]+)\t\t[-0-9.s]*\t(.*)/\1 \2 \3/p' <<<"$ALL_META")

# compute final geometry

# $1: src dimension to keep
# $2: src dimension to scale rel. PAR
# $3 $4: PAR num/den
# $5 $6: resizing num/den
# &1: "<adjusted $1> <adjusted $2> <resized+rounded $1> <scaled+resized+rounded $2> <$5 for convenience> <$6…>"
function scale_round() {
  local dim1=$(( ($1+4)*$5/$6/16*16 )) # round ¾-to-floor, ¼-to-ceil
  local dim2=$(( ($2*$3/$4+4)*$5/$6/16*16 )) # round ¾-to-floor, ¼-to-ceil
  local adj1=$(( dim1*$6/$5 ))
  local adj2=$(( dim2*$6*$4/$5/$3 ))
  printf '%d %d %d %d %d %d' $adj1 $adj2 $dim1 $dim2 $5 $6
}

declare -A SRC_RAW SRC_WHLT SRC_WH2WH CATEG_TIME

# → compute optimal crop and sizes, to target a pixel aspect ratio of 1:1, with width and height being multiples of 16
for tnum in $TITLE_LIST; do
  read -r raww rawh parn pard cropw croph cropl cropt interl id < <(sed -nr "
      s/^V-$tnum\t(\t[^\t]*){2}\t([0-9]+)x([0-9]+)/\2 \3/; T
      s#(\t[^\t]*){2}\t([0-9]+)/([0-9]+)\t([0-9]+):([0-9]+):([0-9]+):([0-9]+)\t([^\t]*)\t([^\t]*)# \2 \3 \4 \5 \6 \7 \8 \9#p
    " <<<"$ALL_META")
  if [ $parn -gt $pard ]; then # stretch horizontally
    read -r srch srcw finalh finalw scalen scaled < <(scale_round $croph $cropw $parn $pard 1 1)
    if [ $finalw -gt $MAXW ] || [ $finalh -gt $MAXH ]; then
      if [ $((100*finalw/MAXW)) -gt $((100*finalh/MAXH)) ]; then
        read -r srch srcw finalh finalw scalen scaled < <(scale_round $croph $cropw $parn $pard $MAXW $finalw)
      else
        read -r srch srcw finalh finalw scalen scaled < <(scale_round $croph $cropw $parn $pard $MAXH $finalh)
      fi
    fi
  else                         # stretch vertically
    read -r srcw srch finalw finalh scalen scaled < <(scale_round $cropw $croph $pard $parn 1 1)
    if [ $finalw -gt $MAXW ] || [ $finalh -gt $MAXH ]; then
      if [ $((100*finalw/MAXW)) -gt $((100*finalh/MAXH)) ]; then
        read -r srcw srch finalw finalh scalen scaled < <(scale_round $cropw $croph $pard $parn $MAXW $finalw)
      else
        read -r srcw srch finalw finalh scalen scaled < <(scale_round $cropw $croph $pard $parn $MAXH $finalh)
      fi
    fi
  fi
  [ $srcw -le $raww ] || srcw=$raww
  [ $srch -le $rawh ] || srch=$rawh
  SRC_WHLT[$tnum]="$cropw $croph $cropl $cropt"
  SRC_WH2WH[$tnum]="$srcw $srch $finalw $finalh $scalen $scaled"
  SRC_RAW[$tnum]="$raww $rawh $parn $pard $interl $id"
  CATEG_TIME["$finalw $finalh"]=$(bc -q <<<"${CATEG_TIME["$finalw $finalh"]:-0}+${VSECONDS[$tnum]}")
done

# → identify the width and height most often used in all chapters
read -r maxw maxh x < <(
  for wh in "${!CATEG_TIME[@]}"; do printf '%s %d\n' "$wh" ${CATEG_TIME[$wh]}; done | sort -k3,3nr -k1,1nr -k2,2nr | head -n 1)
unset CATEG_TIME

# → rescale each title according to the chosen width and height
for tnum in $TITLE_LIST; do
  read -r sw sh fw fh scn scd <<<"${SRC_WH2WH[$tnum]}"
  read -r rw rh parn pard x <<<"${SRC_RAW[$tnum]}"
  # already OK ⇒ skip
  if { [ $fw -eq $maxw ] || [ $fh -eq $maxh ]; } && [ $fw -le $maxw ] && [ $fh -le $maxh ]; then
    continue
  fi
  # else scale final geometry
  if [ $parn -gt $pard ]; then
    read -r newsrch newsrcw newh neww newscalen newscaled < <(scale_round $sh $sw $parn $pard $((scn*maxw)) $((scd*fw)))
    if [ $newh -gt $maxh ]; then
      read -r newsrch newsrcw newh neww newscalen newscaled < <(scale_round $sh $sw $parn $pard $((scn*maxh)) $((scd*fh)))
    fi
  else
    read -r newsrcw newsrch neww newh newscalen newscaled < <(scale_round $sw $sh $pard $parn $((scn*maxw)) $((scd*fw)))
    if [ $newh -gt $maxh ]; then
      read -r newsrcw newsrch neww newh newscalen newscaled < <(scale_round $sw $sh $parn $pard $((scn*maxh)) $((scd*fh)))
    fi
  fi
  SRC_WH2WH[$tnum]="$newsrcw $newsrch $neww $newh $newscalen $newscaled"
done

# compute final audio streams

declare -A SRC_AUDIO CATEG_TIME

# → categorize all audio streams
for tch in "${!CSECONDS[@]}"; do
  allaudio="$(awk -F$'\t' -vT=${tch%.*} -vC=${tch#*.} '$1=="A-" T && $2==C{print T "." C, $3, $14, $5, $8, $9, $12, $4, $6, $7, $11, $13}' <<<"$ALL_META")"
  # (technical info:) <1:title.chapter number> <2:stream number> <3:id> <4:codec> <5:Hz> <6:bit/s> <7:secondary?>
  # (category:) <8:lang> <9:channels count> <10:layout> <11:commentary?> <12:visu.impaired?> <13:priority>
  SRC_AUDIO[$tch]="$(
    for lng in ${ALL_LANGS//,/ }; do
      awk -vL=$lng '$8==L{print}' <<<"$allaudio" \
      | sort -k11,11 -k12,12 -k7,7 -k9,9nr -k10,10 -k2,2n \
      | awk '($9 $10 $11 $12)!=ref{ref=($9 $10 $11 $12);c=0};{c++; print $0, c}'
    done
  )"
  while read -r x x x cod freq bps x lng cnt lay cmt blind prio; do
    x="$cod $freq $bps $lng $cnt $lay $cmt $blind $prio"
    CATEG_TIME["$x"]=$(bc -q <<<"${CATEG_TIME["$x"]:-0}+${CSECONDS[$tch]%.*}")
  done <<<"${SRC_AUDIO[$tch]}"
done

# → identify for each category the audio settings most often used in all chapters, with preferred language on top
declare -A expo=(["1"]='' ["2"]=² ["3"]=³ ["4"]=⁴ ["5"]=⁵ ["6"]=⁶ ["7"]=⁷ ["8"]=⁸ ["9"]=⁹)
allaudio="$(export IFS=$'\n'; echo "${SRC_AUDIO[*]}")"
# (technical info:) <1:stream number> <2:codec> <3:Hz> <4:bit/s>
# (category:) <5:lang> <6:channels count> <7:layout> <8:commentary?> <9:visu.impaired?> <10:priority> <11:name>
MKV_AUDIO="$(
  for lng in ${ALL_LANGS//,/ }; do
    while read -r x x x cod freq bps x lng cnt lay cmt blind prio; do
      x="$cod $freq $bps $lng $cnt $lay $cmt $blind $prio"
      name="${lng^^} ${cnt}🕩 $lay"
      [ "$blind" == true ] && name+=" 🙈"
      [ "$cmt" == true ] && name+=" 🗩"
      echo "${CATEG_TIME["$x"]} $cod $freq $bps $lng $cnt $lay $cmt $blind $prio $name${expo[$prio]}"
    done < <(awk -vL=$lng '$8==L{print}' <<<"$allaudio") \
    | sort -k8,8 -k9,9 -k6,6nr -k7,7 -k10,10n -k1,1nr \
    | uniq -f5
  done \
  | nl -nln -s' ' -w1 | cut -d' ' -f1,3-
)"
unset expo allaudio
unset CATEG_TIME

# compute final subtitle streams

declare -A SRC_SUB

# → categorize all subtitle streams
for tch in "${!CSECONDS[@]}"; do
  allsubs="$(awk -F$'\t' -vT=${tch%.*} -vC=${tch#*.} '$1=="S-" T && $2==C{print $3, $10, $8, $5, $4, $7, $9}' <<<"$ALL_META")"
  # (technical info:) <1:stream number> <2:id> <3:forced?>
  # (category:) <4:codec> <5:lang> <6:commentary?> <7:hear.impaired?> <8:priority>
  SRC_SUB[$tch]="$(
    for lng in ${ALL_LANGS//,/ }; do
      awk -vL=$lng '$5==L{print}' <<<"$allsubs" | sort -k6,6 -k7,7 -k3,3r -k4,4 -k1,1n | awk '($4 $6 $7)!=ref{ref=($4 $6 $7);c=0};{c++; print $0, c}'
    done
  )"
done

# → identify final subtitle streams, with preferred language on top
declare -A expo=(["1"]='' ["2"]=² ["3"]=³ ["4"]=⁴ ["5"]=⁵ ["6"]=⁶ ["7"]=⁷ ["8"]=⁸ ["9"]=⁹)
allsubs="$(export IFS=$'\n'; echo "${SRC_SUB[*]}")"
# (technical info:) <1:stream number>
# (category:) <2:codec> <3:lang> <4:commentary?> <5:hear.impaired?> <6:priority> <7:name>
MKV_SUB="$(
  for lng in ${ALL_LANGS//,/ }; do
    while read -r x x x cod lng cmt deaf prio; do
      name="${lng^^}"
      [ "$deaf" == true ] && name+=" 🙉"
      [ "$cmt" == true ] && name+=" 🗩"
      echo "$cod $lng $cmt $deaf $prio $name${expo[$prio]}"
    done < <(awk -vL=$lng '$5==L{print}' <<<"$allsubs") \
    | sort -k3,3 -k4,4 -k1,1 -k5,5n \
    | uniq
  done \
  | nl -nln -s' ' -w1
)"
unset expo

# generate per-chapter timeline snapshots

[ -d "$BUILD/.tsnap" ] || mkdir "$BUILD/.tsnap"
while read -r tnum chnum seconds; do
  if ! [ -f "$BUILD/t.$tnum.ch.$chnum.png" ]; then
    audio=( $(awk -vOFS=, '{print $8, $9, $10, $11, $12, $13}' <<<"${SRC_AUDIO[$tnum.$chnum]}") )
    subs=( $(awk -vOFS=, '{print $4, $5, $6, $7, $8}' <<<"${SRC_SUB[$tnum.$chnum]}") )
    gridsize=$(($(bc -lq <<<"sqrt($seconds/5)" | sed 's/\.0*$//;s/\..*/+1/')))
    log_and_run "Make timeline snapshot of track $tnum chapter $chnum" \
    ffmpeg -hide_banner -f dvdvideo -trim false -title $tnum -chapter_start $chnum -chapter_end $chnum -i "$BUILD/.iso" \
      -map 0:v:0 -frames:v 1 -filter:v \
      "select='bitor(gte(t-prev_selected_t,5),isnan(prev_selected_t))',scale=w=256:h=144:force_original_aspect_ratio=decrease,tile=${gridsize}x${gridsize}" \
      -fps_mode passthrough "$BUILD/.tsnap/t.$tnum.ch.$chnum.png" </dev/null
    log_and_run "Annotate timeline snapshot of track $tnum chapter $chnum with audio and subtitles metadata" \
    convert -fill red -gravity SouthEast -annotate +0+0 \
      "$(export IFS=+; echo "${audio[*]}")|$(export IFS=+; echo "${subs[*]}")" \
      "$BUILD/.tsnap/t.$tnum.ch.$chnum.png" "$BUILD/t.$tnum.ch.$chnum.png" </dev/null
    unset audio subs
  fi
done < <(
  # restrict to chapters having common durations
  for tch in ${!CSECONDS[@]}; do printf '%s %s\n' "${tch/./ }" ${CSECONDS[$tch]}; done \
  | sort -k3,3nr | uniq -f2 -D # keep only duplicates in the duration column
)

# detect duplicates

function islongfilm() {
  [ ${VSECONDS[$1]} -ge 4200 ] # [70min–…]
  echo $?
}

function isepisode() {
  [ ${VSECONDS[$1]} -ge 900 ] # [15min–…]
  echo $?
}

declare -A STREAMS_IDS

for tch in "${!CSECONDS[@]}"; do
  strid=
  while read -r x x x x x x x lng x lay cmt blind prio; do
    strid+=",${lng}${prio}:${lay}(${cmt}/${blind})"
  done <<<"${SRC_AUDIO[$tch]}"
  while read -r x x x cod lng cmt deaf prio; do
    strid+=",${lng}${prio}:${cod}(${cmt}/${deaf})"
  done <<<"${SRC_SUB[$tch]}"
  STREAMS_IDS[$tch]="$strid"
done

if ! [ -f "$BUILD/.duplicates" ]; then
  log_and_run -o "Detect duplicates among all chapters of all titles" \
  czkawka_cli image -f "$BUILD/.raw_dedup" -s High -m 1 -R -d "$BUILD/" \
  | sed -rn 's#^"?/.*/t\.([0-9]+)\.ch\.([0-9]+)\.png.*#\1 \2#p' \
  | while read -r tnum1 chnum1; read -r tnum2 chnum2; do
    if [ "${STREAMS_IDS[$tnum1.$chnum1]}" == "${STREAMS_IDS[$tnum2.$chnum2]}" ]; then
      # long film first, else episode first, else lower title number first, else lower chapter number first
      printf '%d %d %d %d %d %d\n%d %d %d %d %d %d\n' \
        $(islongfilm $tnum1) $(isepisode $tnum1) $tnum1 $tnum2 $chnum1 $chnum2 \
        $(islongfilm $tnum2) $(isepisode $tnum2) $tnum2 $tnum1 $chnum2 $chnum1 \
      | sort -k1,1n -k2,2n -k3,3n -k4,4n -k5,5n | head -n 1 | awk '{print $3, $5, $4, $6}'
    fi
  done >"$BUILD/.duplicates"
fi
unset STREAMS_IDS

declare -A SUBST

while read -r t1 ch1 t2 ch2; do
  SUBST[$t2.$ch2]="$t1.$ch1"
done <"$BUILD/.duplicates"

# rip chapters and fill-in missing streams

function effectivePlayTime() {
  ffmpeg -hide_banner "$@" -f null /dev/null </dev/null \
  |& grep -o 'time=[^ ]*' | tail -n 1 | tr '=.' : | awk -F: '{printf("%d.%s",3600*$2+60*$3+$4,$5)}'
}

function within1s() {
  [ -z "$(bc -q <<<"(${1:-0}-${2:-0})^2" | cut -d. -f1)" ]
}

declare -A USED_A USED_S

# $1,2,3: tnum, chnum, chtime
# $5, $6: (optional) alt. raw input, id-to-alt. stream number mapping ('id num\nid num\n…')
function ripchapter() {
  local tnum=$1 chnum=$2 chtime=$3 altinput="$4" altmap="$5"
  local -A blank
  local -a ffcmd ffinput ffmap ffenc
  if [ -n "$altinput" ]; then
    ffcmd=(ffmpeg -hide_banner)
    ffinput=(-i "$altinput")
  else
    ffcmd=(ffmpeg -hide_banner -f dvdvideo -trim false -title $tnum -chapter_start $chnum -chapter_end $chnum)
    ffinput=(-i "$BUILD/.iso")
  fi
  ffmap=(-map_chapters -1)
  ffenc=()

  # video stream
  if [ -n "${SRC_VIDEO[$tnum]}" ]; then
    read -r x x x x interl id <<<"${SRC_RAW[$tnum]}"
    num=${SRC_VIDEO[$tnum]}
    if [ -n "$altinput" ]; then
      num=$(awk -vI="$id" '$1==I{print $2}' <<<"$altmap")
      if [ -z "$num" ]; then
        printf 'ERROR: COULD NOT RETRIEVE VIDEO OF TITLE %d CHAPTER %d\n' $tnum $chnum >&2
        exit 1
      fi
    fi
    read -r sw sh fw fh x <<<"${SRC_WH2WH[$tnum]}"
    read -r x x cropl cropt <<<"${SRC_WHLT[$tnum]}"

    ffmap+=(-map 0:$num)
    filter=
    if [ "$interl" == true ]; then
      filter+=",bwdif"
    fi
    filter+=",crop=x=$cropl:y=$cropt:w=$((sw/2*2)):h=$((sh/2*2)):exact=1"
    if [ $sw -ne $fw ] || [ $sh -ne $fh ]; then
      ffenc+=(-sws_flags lanczos+accurate_rnd)
      filter+=",scale=w=$fw:h=$fh:force_original_aspect_ratio=disable"
    fi
    filter+=',setsar=1/1'
    if [ $fw -ne $maxw ] || [ $fh -ne $maxh ]; then
      filter+=",pad=w=$maxw:h=$maxh:x=-1:y=-1"
    fi
    ffenc+=(-filter:v:0 "${filter:1}" -enc_time_base:v:0 demux)
    ffenc+=(-c:v:0 libx265 -x265-params:v:0 profile=main10:preset=slower:crf=18:sar=1:videoformat=${VFORMAT:-pal}:rc-lookahead=120:bframes=12:ref=6:subme=7:aq-mode=3)
  fi

  # audio streams
  while read -r mkvnum ffcod mkvfreq mkvrate lng count ffchanlay cmt blind prio name; do if [ -n "$mkvnum" ]; then
    read -r x num id cod freq rate x < <(
      awk -vL=$lng -vC=$count -vY=$ffchanlay -vM=$cmt -vB=$blind -vP=$prio '$8==L && $9==C && $10==Y && $11==M && $12==B && $13==P{print}' <<<"${SRC_AUDIO[$tnum.$chnum]}")
    [ -n "$altinput" ] && num=$(awk -vI="$id" '$1==I{print $2}' <<<"$altmap")

    if [ -z "$num" ]; then
      # no such stream in this title
      f="$BUILD/tmp.empty_a.$ffchanlay.$mkvfreq.$mkvrate.$chtime.$ffcod"
      if ! [ -f "$f" ] && ! [ -f "$BUILD/t.$tnum.ch.$chnum.mkv" ]; then
        log_and_run "For title $tnum chapter $chnum, create missing audio stream for ${chtime}s of ${mkvnum}: ${name}" \
        ffmpeg -hide_banner -strict -2 -lavfi anullsrc=channel_layout="${ffchanlay}":sample_rate=${mkvfreq}:duration=${chtime} -c:a ${ffcod} -b:a ${mkvrate} "$f" </dev/null
      fi
      if [ -z "${blank["$f"]}" ]; then
        ffinput+=(-i "$f")
        blank["$f"]=$((${#blank[*]}+1))
      fi
      ffmap+=(-map ${blank["$f"]}:a:0)
      ffenc+=(-c:a:$((mkvnum-1)) copy)
    elif [ "$cod,$freq,$rate" == "$ffcod,$mkvfreq,$mkvrate" ]; then
      # exact match in this title
      grep -qF " $mkvnum " <<<"${USED_A[$tnum]} " || USED_A[$tnum]="${USED_A[$tnum]} $mkvnum"
      ffmap+=(-map 0:$num)
      ffenc+=(-c:a:$((mkvnum-1)) copy -map_metadata:s:a:$((mkvnum-1)) 0:s:$num)
    else
      # steam has different characteristics
      grep -qF " $mkvnum " <<<"${USED_A[$tnum]} " || USED_A[$tnum]="${USED_A[$tnum]} $mkvnum"
      ffmap+=(-map 0:$num)
      ffenc+=(-c:a:$((mkvnum-1)) $ffcod -ar:a:$((mkvnum-1)) $mkvfreq -b:a:$((mkvnum-1)) $mkvrate -map_metadata:s:a:$((mkvnum-1)) 0:s:$num)
    fi
    ffenc+=(-metadata:s:a:$((mkvnum-1)) language=$lng -metadata:s:a:$((mkvnum-1)) title="$mkvnum: $name")
  fi; done <<<"$MKV_AUDIO"

  # subtitle streams
  while read -r mkvnum ffcod lng cmt deaf prio name; do if [ -n "$mkvnum" ]; then
    read -r num id x < <(
      awk -vC=$ffcod -vL=$lng -vM=$cmt -vD=$deaf -vP=$prio '$4==C && $5==L && $6==M && $7==D && $8==P{print}' <<<"${SRC_SUB[$tnum.$chnum]}")
    [ -n "$altinput" ] && num=$(awk -vI="$id" '$1==I{print $2}' <<<"$altmap")

    if [ -n "$num" ]; then
      # exact match in this title
      grep -qF " $mkvnum " <<<"${USED_S[$tnum]} " || USED_S[$tnum]="${USED_S[$tnum]} $mkvnum"
      ffmap+=(-map 0:$num)
      ffenc+=(-c:s:$((mkvnum-1)) copy -map_metadata:s:s:$((mkvnum-1)) 0:s:$num)
    elif [ "$ffcod" != dvd_subtitle ]; then
      # unsupported type!
      printf 'ERROR: UNSUPPORTED MISSING SUBTITLE STREAM FOR TITLE %d CHAPTER %d OF TYPE %s\n' $tnum $chnum "$ffcod" >&2
      exit 1
    else
      # no such stream in this title
      f="$BUILD/tmp.empty_s.${VFORMAT:-pal}.$chtime.mpeg2"
      if ! [ -f "$f" ] && ! [ -f "$BUILD/t.$tnum.ch.$chnum.mkv" ]; then
        if ! [ -f "$BUILD/tmp.empty_pixel.png" ]; then
          log_and_run "Generate a transparent pixel for use as an empty subtitle stream" \
          convert -size 1x1 'xc:rgba(0,0,0,0)' "$BUILD/tmp.empty_pixel.png"
        fi
        log_and_run -o "Generate XML description for ${chtime}s of empty subtitle" \
        echo "<subpictures format=\"$([ "$VFORMAT" == ntsc ] && echo NTSC || echo PAL)\"><stream><spu start=\"00:00:00.00\" end=\"$(date -u -d@${chtime} +%T.${chtime#*.})\" image=\"$BUILD/tmp.empty_pixel.png\"/></stream></subpictures>" >"$BUILD/tmp.empty_s.xml"
        log_and_run -o "For title $tnum chapter $chnum, create missing subtitle stream for ${chtime}s of ${mkvnum}: ${name}" \
        spumux -m dvd --nomux --nodvdauthor-data "$BUILD/tmp.empty_s.xml" </dev/null >"$f"
      fi
      if [ -z "${blank["$f"]}" ]; then
        ffinput+=(-i "$f")
        blank["$f"]=$((${#blank[*]}+1))
      fi
      ffmap+=(-map ${blank["$f"]}:s:0)
      ffenc+=(-c:s:$((mkvnum-1)) copy)
    fi
    ffenc+=(-metadata:s:s:$((mkvnum-1)) language=$lng -metadata:s:s:$((mkvnum-1)) title="$mkvnum: $name")
  fi; done <<<"$MKV_SUB"

  # encode
  if ! [ -f "$BUILD/t.$tnum.ch.$chnum.mkv" ]; then (
    if [ -n "$altinput" ]; then
      log_and_run "Adapt title $tnum chapter $chnum to final MKV streams (from raw extract)" \
      "${ffcmd[@]}" "${ffinput[@]}" "${ffmap[@]}" "${ffenc[@]}" "$BUILD/t.$tnum.ch.$chnum.mkv" </dev/null \
      && exit 0
      printf 'ERROR: UNABLE TO RIP AND ENCODE TITLE %d CHAPTER %d\n' $tnum $chnum >&2
      exit 1
    fi
    duration=$(effectivePlayTime -f dvdvideo -preindex 1 -trim false -title $tnum -chapter_start $chnum -chapter_end $chnum -i "$BUILD/.iso")
    if within1s $chtime $duration; then
      log_and_run "Adapt title $tnum chapter $chnum to final MKV streams" \
      "${ffcmd[@]}" -preindex 1 "${ffinput[@]}" "${ffmap[@]}" "${ffenc[@]}" "$BUILD/t.$tnum.ch.$chnum.mkv" </dev/null \
      && exit 0
    fi
    duration=$(effectivePlayTime -f dvdvideo -trim false -title $tnum -chapter_start $chnum -chapter_end $chnum -i "$BUILD/.iso")
    if within1s $chtime $duration; then
      log_and_run "Adapt title $tnum chapter $chnum to final MKV streams (degraded)" \
      "${ffcmd[@]}" "${ffinput[@]}" "${ffmap[@]}" "${ffenc[@]}" "$BUILD/t.$tnum.ch.$chnum.mkv" </dev/null \
      && exit 0
    fi
    if ! [ -f "$BUILD/t.$tnum.ch.$chnum.mpeg" ]; then
      log_and_run "Extract title $tnum chapter $chnum raw data" \
      mplayer dvd://${tnum} -chapter ${chnum}-${chnum} -dumpstream -dumpfile "$BUILD/t.$tnum.ch.$chnum.mpeg" -dvd-device "$BUILD/.iso" </dev/null
    fi
    duration=$(effectivePlayTime -i "$BUILD/t.$tnum.ch.$chnum.mpeg")
    if within1s $chtime $duration || [ $(tail -n 1 <<<"${CHAPTERS[$tnum]}") == $chnum ]; then
      altmap="$(ffprobe -hide_banner -output_format json -show_streams -analyzeduration "$(bc -q <<<"$duration*1000000" | cut -d. -f1)" \
        "$BUILD/t.$tnum.ch.$chnum.mpeg" 2>/dev/null | jq -r '.streams[] |"\(.id) \(.index)"')"
      ripchapter $tnum $chnum $chtime "$BUILD/t.$tnum.ch.$chnum.mpeg" "$altmap" \
      && exit 0
    fi
    if [ ${duration%.*}0 -lt 600 ]; then # <60s
      printf '\nWARNING: Cannot read the exact length of title %d chapter %d (%s over %s); skipping…\n' $tnum $chnum $duration $chtime >&2
      exit 0 # don’t rip; it is probably a subtitle/menu chapter
    fi
    printf '\nERROR: CANNOT READ THE EXACT LENGTH OF TITLE %d CHAPTER %d\n' $tnum $chnum >&2
    exit 1
  ); fi \
  && rm -f "$BUILD/tmp.empty_s.xml" "$BUILD/tmp.empty_pixel.png"
}

while read -r tnum chnum chtime; do
  [ -z "${SUBST[$tnum.$chnum]}" ] || continue
  [ ${chtime:0:1} == . ] && chtime=0$chtime
  ripchapter $tnum $chnum $chtime || exit 1
  if [ -f "$BUILD/t.$tnum.ch.$chnum.mpeg" ] && [ $(tail -n 1 <<<"${CHAPTERS[$tnum]}") == $chnum ]; then
    CSECONDS[$tnum.$chnum]=$(effectivePlayTime -i "$BUILD/t.$tnum.ch.$chnum.mkv")
  fi
done < <(for tch in ${!CSECONDS[@]}; do printf '%s %s\n' "${tch/./ }" ${CSECONDS[$tch]}; done | sort -k1,1n -k2,2n)
for substid in "${!SUBST[@]}"; do
  tch=$substid
  while [ -n "${SUBST[$tch]}" ]; do tch=${SUBST[$tch]}; done
  [ -n "${USED_A[${substid%.*}]}" ] || USED_A[${substid%.*}]="${USED_A[${tch%.*}]}"
  [ -n "${USED_S[${substid%.*}]}" ] || USED_S[${substid%.*}]="${USED_S[${tch%.*}]}"
done
unset SRC_VIDEO SRC_RAW SRC_WHLT SRC_WH2WH SRC_AUDIO SRC_SUB

# assemble source tracks into end-result tracks

if ! [ -f "$BUILD/.titlelist" ]; then
  tcategs="$(
    for tnum in $TITLE_LIST; do
      printf '%d %d %d %d\n' $(islongfilm $tnum) $(isepisode $tnum) $tnum $(ls -1 "$BUILD"/t.$tnum.ch.*.mkv | wc -l)
    done \
    | grep -v ' 0$'
  )"
  {
    awk '$1==0 {printf("🎥 %d\t%d\n",$3,$3)}' <<<"$tcategs"
    awk '$1==1 && $2==0 {L=L " " $3}; END {if(L!="")printf("📹%s\t%s\n",gensub(" "," ","g",L),gensub(" ","",1,L))}' <<<"$tcategs"
    awk '$1==1 && $2==1 {L=L " " $3}; END {if(L!="")printf("🖼%s\t%s\n",gensub(" "," ","g",L),gensub(" ","",1,L))}' <<<"$tcategs"
    awk '{L=L " " $3}; END {if(L!="")printf("…🎞🎞🎞…\t%s\n",gensub(" ","",1,L))}' <<<"$tcategs"
    {
      if [ $(grep '^1 0 ' <<<"$tcategs" | wc -l) -gt 1 ]; then
        awk '$1==1 && $2==0 {printf("🎞 %d\t%d\n",$3,$3)}' <<<"$tcategs"
      fi
      if [ $(grep '^1 1 ' <<<"$tcategs" | wc -l) -gt 1 ]; then
        awk '$1==1 && $2==1 {printf("🎞 %d\t%d\n",$3,$3)}' <<<"$tcategs"
      fi
    } | sort -t$'\t' -k2,2n
  } >"$BUILD/.titlelist"
  unset tcategs
fi

# fetch chapters’ MKV metadata

for mkv in "$BUILD/"t.*.ch.*.mkv; do
  if ! [ -f "${mkv%mkv}json" ]; then
    log_and_run "Read metadata (length, ids…) for $mkv" \
    mkvmerge -F json -i "$mkv" -r "${mkv%mkv}json"
  fi
done

# compute chapters’ timestamps, UIDs, and substitutions + prepare merging

declare -A TIMESTAMPS
declare -A CHAPTERS_IDS
declare -A UIDS
declare -a MKVMRGARGS=()

default=
strnum=0
while read -r x x x x lng x x cmt blind x; do if [ -n "$lng" ]; then
  strnum=$((strnum+1))
  if [ "$default$cmt$blind${lng,,}" == "falsefalse${DEFAULT_LANG,,}" ]; then
    default=found
    MKVMRGARGS+=(--default-track-flag $strnum:1)
  else
    MKVMRGARGS+=(--default-track-flag $strnum:0)
  fi
  MKVMRGARGS+=(--visual-impaired-flag $strnum:$([ "$blind" == true ] && printf 1 || printf 0))
  MKVMRGARGS+=(--commentary-flag $strnum:$([ "$cmt" == true ] && printf 1 || printf 0))
fi; done <<<"$MKV_AUDIO"
while read -r x x lng cmt deaf x; do if [ -n "$lng" ]; then
  strnum=$((strnum+1))
  if [ "$default$cmt$blind" == "falsefalse" ] && grep -qiF ",${lng,,}," <<<",$ALL_LANGS,"; then
    default=found
    MKVMRGARGS+=(--default-track-flag $strnum:1)
  else
    MKVMRGARGS+=(--default-track-flag $strnum:0)
  fi
  MKVMRGARGS+=(--hearing-impaired-flag $strnum:$([ "$deaf" == true ] && printf 1 || printf 0))
  MKVMRGARGS+=(--commentary-flag $strnum:$([ "$cmt" == true ] && printf 1 || printf 0))
fi; done <<<"$MKV_SUB"
unset default strnum

ref=0
MKVMRGARGS+=('--no-chapters' '--no-global-tags' '[')
while read -r tnum; do
  for chnum in ${CHAPTERS[$tnum]}; do if [ -f "$BUILD/t.$tnum.ch.$chnum.mkv" ]; then
    [ -z "${SUBST[$tnum.$chnum]}" ] || continue
    MKVMRGARGS+=("$BUILD/t.$tnum.ch.$chnum.mkv")
    # compute floating-point seconds
    start=$ref
    stop=$(bc -lq <<<"$ref+$(jq -r '.container.properties | .duration/.timestamp_scale/1000' <"$BUILD/t.$tnum.ch.$chnum.json")")
    ref=$stop
    TIMESTAMPS[$tnum.$chnum]="$start $stop"
    CHAPTERS_IDS[$tnum.$chnum]="$tnum$chnum$(jq -r .container.properties.segment_uid <"$BUILD/t.$tnum.ch.$chnum.json" \
      | tr '[a-f]' '[A-F]' | bc -q <<<"ibase=16;$(cat)" | head -c 8)"
    UIDS[$tnum.$chnum]="$(jq -r .container.properties.segment_uid <"$BUILD/t.$tnum.ch.$chnum.json")"
  fi; done
done < <(cut -d$'\t' -f2 <"$BUILD/.titlelist" | tr ' ' '\n' | awk '{if($0 in seen)next};{print;seen[$0]=1}')
for substid in "${!SUBST[@]}"; do
  tch=$substid
  while [ -n "${SUBST[$tch]}" ]; do tch=${SUBST[$tch]}; done
  TIMESTAMPS[$substid]=${TIMESTAMPS[$tch]}
  CHAPTERS_IDS[$substid]="${substid//.}$(jq -r .container.properties.segment_uid <"$BUILD/t.${tch//./.ch.}.json" \
    | tr '[a-f]' '[A-F]' | bc -q <<<"ibase=16;$(cat)" | head -c 8)"
  UIDS[$substid]="${UIDS[$tch]}"
done
MKVMRGARGS+=(']')
unset ref

# generate MKV editions XML file

function formatTimestamp() {
  if [ $1 == 0 ]; then
    echo '00:00:00.000'
  else
    date -u -d@${1%.*} +%T.${1#*.}
  fi
}

if ! [ -f "$BUILD/mkv-editions.xml" ]; then
{
  cat <<ENDOFXML
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE Chapters SYSTEM "matroskachapters.dtd">
<Chapters>
ENDOFXML

  edflagdefault=1
  titlenum=1
  while IFS=$'\t' read -r label titlelist; do

    cat <<ENDOFXML
  <EditionEntry>
    <EditionFlagHidden>0</EditionFlagHidden>
    <EditionFlagDefault>$edflagdefault</EditionFlagDefault>
    <EditionFlagOrdered>1</EditionFlagOrdered>
    <EditionDisplay>
      <EditionString>$(printf '%03d %s' $titlenum "$label")</EditionString>
    </EditionDisplay>
ENDOFXML

    chflaghidden=0
    for tnum in $titlelist; do
      for chnum in ${CHAPTERS[$tnum]}; do if [ -f "$BUILD/t.$tnum.ch.$chnum.mkv" ]; then

        cat <<ENDOFXML
    <ChapterAtom>
      <ChapterUID>${CHAPTERS_IDS[$tnum.$chnum]}</ChapterUID>
      <ChapterTimeStart>$(formatTimestamp ${TIMESTAMPS[$tnum.$chnum]% *})</ChapterTimeStart>
      <ChapterTimeEnd>$(formatTimestamp ${TIMESTAMPS[$tnum.$chnum]#* })</ChapterTimeEnd>
      <ChapterFlagHidden>$chflaghidden</ChapterFlagHidden>
      <ChapterFlagEnabled>1</ChapterFlagEnabled>
      <!-- ChapterSegmentUID format="hex">${UIDS[$tnum.$chnum]}</ChapterSegmentUID -->
      <ChapterDisplay>
        <ChapterString>$tnum.$chnum (A${USED_A[$tnum]:- ∅} - S${USED_S[$tnum]:- ∅}) — ${CHNAMES[$tnum.$chnum]}</ChapterString>
      </ChapterDisplay>
    </ChapterAtom>
ENDOFXML

        if [ "${titlelist/ }" != "$titlelist" ]; then
          chflaghidden=1
        fi
      fi; done
      chflaghidden=0
    done

    cat <<ENDOFXML
  </EditionEntry>
ENDOFXML

    edflagdefault=0
    titlenum=$((titlenum+1))
  done <"$BUILD/.titlelist"

  cat <<ENDOFXML
</Chapters>
ENDOFXML

} >"$BUILD/mkv-editions.xml"
fi

# merge all chapters of all titles

if ! [ -f "$TARGET" ]; then
  log_and_run 'Build final MKV file' \
  mkvmerge --attachment-description 'MKV Ordered-Chapters Editions' --attachment-mime-type text/xml --attach-file "$BUILD/mkv-editions.xml" \
    --disable-track-statistics-tags --append-mode file --chapters "$BUILD/mkv-editions.xml" -o "$TARGET" "${MKVMRGARGS[@]}"
fi

if [ $? -eq 0 ] && [ -z "$KEEP_ALL" ]; then
  rm -rf "$BUILD"
fi

# FIXME: HOW TO DETECT USELESS SUBTITLES? E.G.
#    3|fra,VOBSUB:bitmap,false,false,false   Francais (Wide Screen) [VOBSUB] → Normal
#    5|fra,VOBSUB:bitmap,false,false,false   Francais (Wide Screen) [VOBSUB] → ??
#    6|fra,VOBSUB:bitmap,false,false,false   Francais (Wide Screen) [VOBSUB] → Commentary
# ⇓
#   (about 4min for full movie for each subtitle track)
# $ HandBrakeCLI -t 1 -e x265_10bit --encoder-profile main10 -q 1 --vfr -X 128 -Y 96 -a none -s 3 --subtitle-burned=none -i .iso -o sub3.mkv
# $ HandBrakeCLI -t 1 -e x265_10bit --encoder-profile main10 -q 1 --vfr -X 128 -Y 96 -a none -s 5 --subtitle-burned=none -i .iso -o sub5.mkv
# $ HandBrakeCLI -t 1 -e x265_10bit --encoder-profile main10 -q 1 --vfr -X 128 -Y 96 -a none -s 6 --subtitle-burned=none -i .iso -o sub6.mkv
#   (instantaneous)
# $ mkvextract sub3.mkv tracks --raw 1:sub3.sub
# $ mkvextract sub5.mkv tracks --raw 1:sub5.sub
# $ mkvextract sub6.mkv tracks --raw 1:sub6.sub
# $ ls -l sub*
# -rw-r--r-- 1 yves yves 287386078  8 mars  18:54 sub3.mkv
# -rw-r--r-- 1 yves yves   1058252  8 mars  19:35 sub3.sub
# -rw-r--r-- 1 yves yves 286304465  8 mars  19:07 sub5.mkv
# -rw-r--r-- 1 yves yves      2268  8 mars  19:35 sub5.sub
# -rw-r--r-- 1 yves yves 289029155  8 mars  19:12 sub6.mkv
# -rw-r--r-- 1 yves yves   2684952  8 mars  19:35 sub6.sub
